/***********************************************************************
 * SCIDYNE Corporation, Pembroke MA USA
 * Program_MCP4728_Address.ino
 * Version 1.0   08-13-2023
 * By Mark Durgin
 *
 * Utility to program new I2C address of DAC chip
 * 
 * Set the pins you will use for the I2C bus SDA and SCL signals
 * and the IO line used for the LDAC signal
 * 
 * Copyright 2023 SCIDYNE Corporation and its affiliates.  All Rights Reserved.
 * Released under MIT License. See license.txt file or visit https://opensource.org/licenses/MIT
 *
 */

/****************************************************** 
 * IMPORTANT! - MUST SET THESE PARAMETERS BEFORE RUNNING PROGRAM
 */
uint8_t SDApin = 18;           // Sets the pin used for SDA signal:  MKR Zero=11;  UNO=18;  MEGA=20;  DUE=(70u);
uint8_t SCLpin = 19;           // Sets the pin used for SCL signal:  MKR Zero=12;  UNO=19;  MEGA=21;  DUE=(71u);
uint8_t LDACpin = 7;           // Sets the pin used to drive the LDAC of MCP4728, pulled LOW on PCB so drive High when needed
uint8_t I2C_OldAddress = 0x60; // Existing I2C Address of DAC being changed, valid addresses are 0x60-0x67  
uint8_t I2C_NewAddress = 0x62; // New I2C Address of DAC being changed, valid addresses are 0x60-0x67


/******************************************************
 * Global Constants and Variables
 */
#define COMPORT   Serial      // Set which Serial port will be used: UNO, MEGA use Serial, DUE may use Serial1
#define I2Cdelay  1           // Set I2C inter-operation delay in milli-Seconds
#define SUCCESS   1           // Success state
#define FAILURE   !SUCCESS    // Failure state

enum sequence {   // Steps in the programming process
  do_byte1 = 1,   // Start list at #1
  do_byte2,
  do_byte3,
  do_byte4,
  completed
};

/* 
 * This union defines 8-bits which can be accessed multiple ways; as an unsigned 8-bit character,
 * or as specific bit fields matching the format of the MCP4728 change address command. 
 * See MCP4728 data sheet for details
 */
union  versaByte {
   // ... Access as a byte
   uint8_t  byteVal;

   // ... Access as individual fields organized as byte 1
   struct {
      uint8_t rw      : 1;   // Bit0:      Read = 1, Write = 0
      uint8_t addr    : 3;   // Bit[3:1]   3-Bit DAC address 
      uint8_t devcode : 4;   // Bit[7:4]   Device Code, MCP4728 always 0b1100
   } form1;
   
   // ... Access as individual fields organized as bytes 3-4
   struct {
      uint8_t rw      : 1;   // Bit0:      Read = 1, Write = 0
      uint8_t y       : 1;   // Bit1:      Bit1
      uint8_t addr    : 3;   // Bit[4:2]   3-Bit DAC address 
      uint8_t cmd     : 3;   // Bit[7:5]   bit7 = C2, bit6 = C1, bit5 = C0 
   } form2;

};


/******************************************************
 * Subroutine prototypes
 */
bool writeByte_I2C(uint8_t outchar);            // Send an 8-bit value Serially using I2C protocol. Returns SUCCESS or FAILURE
bool writeByte_I2C_with_LDAC(uint8_t outchar);  // Send an 8-bit value Serially using I2C protocol. Asserts LDAC signal. Returns SUCCESS or FAILURE


/******************************************************
 * Execution begins here
 *****************************************************/

/******************************************************
 * setup Runs Once the execution goes to loop
 */
void setup() {

  // Local variables
  uint8_t     sequence;         // Holds the steps during the programming process
  uint8_t     FailFlag;         // Holds status of completed operation i.e., SUCCESS or FAILURE
  versaByte   outByte;          // Used to create the outgoing bytes sent to the DAC
  char        buf[80];          // Buffer for sprintf usage

  // Setup for serial communication to Arduino Serial Monitor
  COMPORT.begin(9600);          // Initialize serial interface
  while (!COMPORT);             // Must wait for some boards  

  // Configure the digital pins as an output and set their initial states
  pinMode(SCLpin, OUTPUT);      // Digital IO pin used as I2C SCL signal
  pinMode(SDApin, OUTPUT);      // Digital IO pin used as I2C SDA signal 
  pinMode(LDACpin, OUTPUT);     // Digital IO pin used to drive DAC LDAC

  digitalWrite(SCLpin, HIGH);   // Set SCL high, I2C Idle state
  digitalWrite(SDApin, HIGH);   // Set SDA high, I2C Idle state
  digitalWrite(LDACpin, LOW);   // Set LDAC Low

  /*
   * Wait for user to start. Observe results using Serial Monitor.
   */

  COMPORT.println(F("********************************************"));
  COMPORT.println(F("* SCIDYNE Corporation                      *"));
  COMPORT.println(F("* Program MULTI-IO-ARD DAC Address Utility *"));
  COMPORT.println(F("* Version 1.0   8-13-2023   Mark Durgin    *"));
  COMPORT.println(F("********************************************"));
  COMPORT.println();

  // Inform user of the operation to be done...
  COMPORT.println(F("DAC will be programmed using these parameters:"));  
  sprintf(buf,"Old DAC Address        : 0x%02X  (Valid Addresses are 0x60 - 0x67)",I2C_OldAddress);
  COMPORT.println(buf);
  sprintf(buf,"New DAC Address        : 0x%02X  (Valid Addresses are 0x60 - 0x67)",I2C_NewAddress);
  COMPORT.println(buf);
  sprintf(buf,"Arduino Digital Pin    : %d",LDACpin); 
  COMPORT.println(buf);
  sprintf(buf,"Arduino SCL Pin        : %d",SCLpin); 
  COMPORT.println(buf);
  sprintf(buf,"Arduino SDA Pin        : %d",SDApin); 
  COMPORT.println(buf);
  COMPORT.println();
  
  // Have user verify setup
  COMPORT.println(F(" - Verify correct I/O VOLTAGE is selected at J9"));
  COMPORT.println(F(" - Enable I2C Pull-Ups by installing SDA and SCL jumpers at J2"));
  COMPORT.println();
  
  sprintf(buf,"If programming DAC0, connect wire between Digital #%d and AP0",LDACpin); 
  COMPORT.println(buf);
  sprintf(buf,"If programming DAC1, connect wire between Digital #%d and AP1",LDACpin); 
  COMPORT.println(buf);
  COMPORT.println();

  // Prompt user to proceed
  COMPORT.println(F("Press RETURN to proceed ... "));
  while (COMPORT.available() == 0);   // Loop till user to presses Carriage Return


  /*
   * Programming begins here
   */

  // Set LDAC High
  digitalWrite(LDACpin, HIGH);    // Set LDAC High
  delay(I2Cdelay);
  
  // Send an I2C start condition
  // by asserting SDA low before SCL
  digitalWrite(SDApin, LOW);
  delay(I2Cdelay); 
  digitalWrite(SCLpin, LOW);
  delay(I2Cdelay); 

  // Initial important variables before programming
  sequence = do_byte1;      // Programming procedure to start at the first byte
  FailFlag = SUCCESS;       // Assume all goes well

  /**************************************************************** 
   * Go through the programming sequence until completed or
   * an error condition has been detected
   */
   
  while ( (sequence != completed) && (FailFlag == SUCCESS) )
  { 
    switch (sequence) {

      // First Byte of sequence
      // Start communicating with intended DAC by sending a write command
      case do_byte1:
        outByte.form1.devcode = 0b1100;                       // Bit [7:4] is device code
        outByte.form1.addr    = ( I2C_OldAddress & 0x07 );    // Bit [3:1] is existing I2C address ( offset from base )
        outByte.form1.rw      = 0;                            // Bit 0     0 = write   
        
        // Send first byte, slave responds with SUCCESS (1) or FAILURE (0)
        ( writeByte_I2C(outByte.byteVal) == SUCCESS ) ? ( sequence = do_byte2 )  :  ( FailFlag = FAILURE );
          
        break;

      // Second Byte of sequence
      case do_byte2:
        outByte.form2.cmd   = 3;                              // Bit [7:5] is Command Type
        outByte.form2.addr  = ( I2C_OldAddress & 0x07 );      // Bit [4:2] is existing I2C address ( offset from base )
        outByte.form2.y     = 0;                              // Bit 1
        outByte.form2.rw    = 1;                              // Bit 0     1 = read
        
        // Send second byte. LDAC gets Asserted.  Slave responds with SUCCESS (1) or FAILURE (0)
        ( writeByte_I2C_with_LDAC(outByte.byteVal) == SUCCESS ) ? ( sequence = do_byte3 )  :  ( FailFlag = FAILURE );
       
      break;


      // Third Byte of sequence
      case do_byte3:
        outByte.form2.cmd    = 3;                             // Bit [7:5] is Command Type
        outByte.form2.addr   = ( I2C_NewAddress & 0x07 );     // Bit [4:2] is existing I2C address ( offset from base )
        outByte.form2.y      = 1;                             // Bit 1
        outByte.form2.rw     = 0;                             // Bit 0     0 = write

        // Send third byte, slave responds with SUCCESS (1) or FAILURE (0)
        ( writeByte_I2C(outByte.byteVal) == SUCCESS ) ? ( sequence = do_byte4 )  :  ( FailFlag = FAILURE );
        
      break;

      
      // Forth Byte of sequence
      case do_byte4:
        outByte.form2.cmd   = 3;                              // Bit [7:5] is Command Type
        outByte.form2.addr  = ( I2C_NewAddress & 0x07 );      // Bit [4:2] is existing I2C address ( offset from base )
        outByte.form2.y     = 1;                              // Bit 1
        outByte.form2.rw    = 1;                              // Bit 0     0 = write

        // Send fourth byte, slave responds with SUCCESS (1) or FAILURE (0)
        ( writeByte_I2C(outByte.byteVal) == SUCCESS ) ? ( sequence = completed )  :  ( FailFlag = FAILURE );
       
       break;

       // Unsupported sequence step
       default:
        FailFlag = FAILURE;      
    
    } // End of switch
  } // End of while loop
  
  
  /*
   * Send an I2C stop condition
   * by asserting SCL high before SDA
   */
  digitalWrite(SDApin, LOW);  // SDA low just in case it's High
  delay(I2Cdelay); 
  digitalWrite(SCLpin, HIGH);
  delay(I2Cdelay);  
  digitalWrite(SDApin, HIGH); // SCL High
  delay(I2Cdelay); 
  
  // Set the LDAC low in case it is still High
  digitalWrite(LDACpin, LOW);     // LDAC Low
  
  // Report results to user
  if (FailFlag == SUCCESS) {
    COMPORT.println(F("DAC Address successfully set to new value."));
  }
  else {
    sprintf(buf,"DAC Address Set procedure FAILED!  I2C ACK missing at Byte number %d",sequence);
    COMPORT.println(buf);
  }
} // End of setup()



/******************************************************
 * Loop Runs continuously
 */
void loop() {
  /*
   * Do Nothing
   */
} // End of loop()



/******************************************************
 * Subroutine begin here
 *****************************************************/

/******************************************************
 * bool writeByte_I2C(uint8_t outchar)
 * 
 * Send an 8-bit value Serially using I2C protocol.
 * Data is sent bit7 first, bit0 last
 * 
 * Enters with SCL LOW, SDA undefined High or Low
 * SDA set to output
 * 
 * Leaves with SCL LOW, SDA HIGH and set as output
 * Returns SUCCESS if acknowledged, FAILURE if not
 */
bool writeByte_I2C(uint8_t outchar){
    
  int shift_count;
  int ack_status;
 
  shift_count = 8;    // 8 data bits per byte

  // Send the byte one bit at a time
  while (shift_count--) {

    // Test Most-Significant-Bit, set SDA line HIGH or LOW accordingly
    (outchar & 0x80) ? ( digitalWrite(SDApin, HIGH) )  :  ( digitalWrite(SDApin, LOW) );
    
    // Generate SCL clock pulse
    delay(I2Cdelay);
    digitalWrite(SCLpin, HIGH);
    delay(I2Cdelay);
    digitalWrite(SCLpin, LOW);
    delay(I2Cdelay);
 
    outchar <<= 1;            // Shift outgoing byte left to test next bit
  }
  
  // Completed sending byte, now allow slave to ACK
  pinMode(SDApin, INPUT);     // Make SDA an input for ACK
  digitalWrite(SCLpin, HIGH); // Take I2C Clock High
  delay(I2Cdelay);
  
  // Read input to determine ACK (0) or NAK (1) received
  ack_status = digitalRead(SDApin);
  
  digitalWrite(SCLpin, LOW);      // Take I2C Clock Low
  delay(I2Cdelay);
  
  digitalWrite(SDApin, HIGH);     // Set output register high before changing DDR
  pinMode(SDApin, OUTPUT);        // Turn SDA back to an output
  digitalWrite(SDApin, HIGH);     // Set output high to exit
  delay(I2Cdelay);
 
  // Return with status of either SUCCESS (1) or Failure (0) depending on Ack
  return ( ack_status ? FAILURE : SUCCESS );

} // End of writeByte_I2C()


/******************************************************
 * bool writeByte_I2C_with_LDAC(uint8_t outchar)
 * 
 * Send an 8-bit value Serially using I2C protocol.
 * Data is sent bit7 first, bit0 last. The LDAC signals
 * is also asserted the appropriate time as required
 * by the MCP4728 chip.
 * 
 * Enters with SCL LOW, SDA undefined High or Low, LDAC High
 * SDA set to output
 * 
 * Leaves with SCL LOW, SDA HIGH, LDAC Low
 * Returns SUCCESS if acknowledged, FAILURE if not
 */
bool writeByte_I2C_with_LDAC(uint8_t outchar){
   
  int shift_count;
  int ack_status;

  shift_count = 8;    // 8 data bits per byte
  
  while (shift_count--) {

    // Test Most-Significant-Bit, set SDA line HIGH or LOW accordingly
    (outchar & 0x80) ? ( digitalWrite(SDApin, HIGH) )  :  ( digitalWrite(SDApin, LOW) );
    
    delay(I2Cdelay);
    digitalWrite(SCLpin, HIGH);
    delay(I2Cdelay);
    digitalWrite(SCLpin, LOW);
    delay(I2Cdelay);
  
    outchar <<= 1;            // Shift outgoing byte left to test next bit
  }
  
  // Drop the LDAC pin to initiate the EEPROM command
  digitalWrite(LDACpin, LOW);     // LDAC Low
  delay(I2Cdelay);
  
  // Completed sending byte, now allow slave to ACK
  pinMode(SDApin, INPUT);       // Make SDA an input for ACK
  digitalWrite(SCLpin, HIGH);   // Take I2C Clock High
  delay(I2Cdelay);
  
  // Read input to determine ACK (0) or NAK (1) received
  ack_status = digitalRead(SDApin);
   
  digitalWrite(SCLpin, LOW);      // Take I2C Clock Low
  delay(I2Cdelay);
  
  digitalWrite(SDApin, HIGH);     // Set output register high before changing DDR
  pinMode(SDApin, OUTPUT);        // Turn SDA back to an output
  digitalWrite(SDApin, HIGH);     // Set output high to exit
  delay(I2Cdelay);
  
  // Return with status of either SUCCESS or FAILURE depending on Ack
  return ( ack_status ? FAILURE : SUCCESS );

} // End of writeByte_I2C_with_LDAC()
